#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <allegro.h>

#include "defs.h"
#include "globals.h"
#include "edit.h"
#include "util.h"
#include "text.h"

/*
   Clears all the data in level_data to defaults.
*/
void NewLevel(LEVEL_DATA *data)
{
   int
      i,
      x,
      y;

   memset(data->name, 0, 41);
   strcpy(data->name, "This level has no name.");
   for(i = 0; i < 64; i++)
   {
      data->reserved[i] = 0;
   }
   for(i = 0; i < 6; i++)
   {
      data->power_pills_pos[i].x = -1;
      data->power_pills_pos[i].y = -1;
   }

   data->ghost_start_pos.x = -1;
   data->ghost_start_pos.y = -1;

   data->pacman_start_pos.x = -1;
   data->pacman_start_pos.y = -1;

   for(y = 0; y < 14; y++)
   {
      for(x = 0; x < 20; x++)
      {
         data->level_data[y][x] = -1;
         data->food_pos[y][x] = 0;
      }
   }
}


/*
   Saves the level.  Returns 1 if successful, 0 if not.
*/
int SaveLevel(char *level, LEVEL_DATA *data)
{
   int
      ret_val = 1;
   FILE
      *fp;

   fp = fopen(level, "wb");
   if(!fp)
   {
      ret_val = 0;
   } else
   {
      if(fwrite(data, sizeof(LEVEL_DATA), 1, fp) != 1)
      {
         ret_val = 0;
      }
      fclose(fp);
   }

   return ret_val;
}


/*
   Loads the level.  Returns 1 if successful, 0 if not.
*/
int LoadLevel(char *level, LEVEL_DATA *data)
{
   int
      ret_val = 1;
   FILE
      *fp;

   fp = fopen(level, "rb");
   if(!fp)
   {
      ret_val = 0;
   } else
   {
      if(fread(data, sizeof(LEVEL_DATA), 1, fp) != 1)
      {
         ret_val = 0;
      }
      fclose(fp);
   }

   return ret_val;
}


/*
   Displays help for the level editor.
*/
void EditHelp(void)
{
   static char
      *editHlp[] = { "Press any key to return...",
                     "\t\t\t\t--------------- Edit Help ---------------",
                     "Keys:",
                     "\tESC     = Quit editing.",
                     "\tENTER   = Places a tile at current position.",
                     "\tDEL     = Delete object at current position.",
                     "\tF1      = This help screen.",
                     "\tF2      = Save level.",
                     "\tF3      = Load level.",
                     "\tF4      = Change level name.",
                     "\tF5      = Pick tile to place.",
                     "\tF6      = Place ghost starting position.",
                     "\tF7      = Place pacman starting position.",
                     "\tF8      = Place food.",
                     "\tF9      = Place power pill.",
                     "\tF10     = Program credits.",
                     NULL};
   int
      y;

   clear(screenBuffer);

   for(y = 1; editHlp[y]; y++)
   {
      NewPrintf(screenBuffer, 0, (y - 1) * 25, editHlp[y]);
   }

   NewPrintf(screenBuffer, 320 - (NewStrlen(editHlp[0]) >> 1),
             460, editHlp[0]);

   UpdateScreen();
   clear_keybuf();
   while(!keypressed()) { }
   clear_keybuf();
}



/*
    The main loop for editing a level.
*/
void EditLevel(void)
{
   int
      done = 0,
      x = 0,
      y = 0,
      tx,
      ty,
      tile = BLOCK1,
      update = 1;
   LEVEL_DATA
      lev_data;
   char
      filename[31];
   BITMAP
      *bmp;

   memset(filename, 0, 31);
   NewLevel(&lev_data);

   while(!done)
   {
      clear_keybuf();

      if(update)
      {
         update = 0;

         /* Update screen */
         clear(screenBuffer);
         /* Display name of level */
         NewPrintf(screenBuffer, 5, 460, lev_data.name);
         RenderTiles(&lev_data, 0, 0, 20, 14);
         RenderFood(&lev_data, 0, 0, 20, 14);
         RenderPowerPills(&lev_data, 0, 0, 20, 14);

         /* Draw ghost start location */
         if(lev_data.ghost_start_pos.x > -1 &&
            lev_data.ghost_start_pos.y > -1)
         {
            bmp = main_data[GHOST1].dat;
            tx = (lev_data.ghost_start_pos.x << 5) + 16 - (bmp->w >> 1);
            ty = (lev_data.ghost_start_pos.y << 5) + 16 - (bmp->h >> 1);
            draw_sprite(screenBuffer, bmp, tx, ty);
         }
         /* Draw pacman start location */
         if(lev_data.pacman_start_pos.x > -1 &&
            lev_data.pacman_start_pos.y > -1)
         {
            bmp = main_data[PAC1_1].dat;
            tx = (lev_data.pacman_start_pos.x << 5) + 16 - (bmp->w >> 1);
            ty = (lev_data.pacman_start_pos.y << 5) + 16 - (bmp->h >> 1);
            draw_sprite(screenBuffer, bmp, tx, ty);
         }
   
         /* Draw current spot */
         rect(screenBuffer, (x << 5), (y << 5), ((x << 5) + 31),
                                 ((y << 5) + 31), 247);
         UpdateScreen();
      }

      /* Check the keys */
      if(key[KEY_ESC])  /* Quit */
      {
         update = 1;
         done = YesNo("Really Quit?  [Y/N]");
      } else if(key[KEY_ENTER])  /* Places a tile */
      {
         update = 1;
         DeleteObject(&lev_data, x, y, DEL_ALL);
         lev_data.level_data[y][x] = tile;
      } else if(key[KEY_DEL])   /* Delete object under cursor */
      {
         update = 1;
         DeleteObject(&lev_data, x, y, DEL_ALL);
      } else if(key[KEY_UP])
      {
         if(y > 0)
         {
            y--;
            update = 1;
         }
      } else if(key[KEY_DOWN])
      {
         if(y < 13)
         {
            y++;
            update = 1;
         }
      } else if(key[KEY_LEFT])
      {
         if(x > 0)
         {
            x--;
            update = 1;
         }
      } else if(key[KEY_RIGHT])
      {
         if(x < 19)
         {
            x++;
            update = 1;
         }
      } else if(key[KEY_F1]) /* Edit help */
      {
         EditHelp();
         update = 1;
      } else if(key[KEY_F2]) /* Save level */
      {
         GetFilename("Save: ", filename);
         SaveLevel(filename, &lev_data);
         update = 1;
      } else if(key[KEY_F3])  /* Load level */
      {
         GetFilename("Load: ", filename);
         LoadLevel(filename, &lev_data);
         update = 1;
      } else if(key[KEY_F4])  /* Give level a name */
      {
         GetLevelName(&lev_data);
         update = 1;
      } else if(key[KEY_F5])  /* Pick a tile to place */
      {
         update = 1;
         tile = PickTile(tile);
      } else if(key[KEY_F6])  /* Place ghost starting pos */
      {
         if(lev_data.level_data[y][x] == -1)
         {
            lev_data.ghost_start_pos.x = x;
            lev_data.ghost_start_pos.y = y;

            if(lev_data.ghost_start_pos.x == lev_data.pacman_start_pos.x &&
               lev_data.ghost_start_pos.y == lev_data.pacman_start_pos.y)
            {
               lev_data.pacman_start_pos.x = -1;
               lev_data.pacman_start_pos.y = -1;
            }
            update = 1;
         }
      } else if(key[KEY_F7])  /* Place pacman starting pos */
      {
         if(lev_data.level_data[y][x] == -1)
         {
            lev_data.pacman_start_pos.x = x;
            lev_data.pacman_start_pos.y = y;

            if(lev_data.ghost_start_pos.x == lev_data.pacman_start_pos.x &&
               lev_data.ghost_start_pos.y == lev_data.pacman_start_pos.y)
            {
               lev_data.ghost_start_pos.x = -1;
               lev_data.ghost_start_pos.y = -1;
            }
            update = 1;
         }
      } else if(key[KEY_F8])  /* Place food */
      {
         update = 1;
         lev_data.food_pos[y][x] = 1;
      } else if(key[KEY_F9])  /* Place power pill */
      {
         update = 1;
         PlacePowerPill(&lev_data, x, y);
      } else if(key[KEY_F10]) /* Credits */
      {
         Credits();
         update = 1;
      }
   }
}

/*
   Gets the name of the level.
*/
void GetLevelName(LEVEL_DATA *level)
{
   Draw3DBox(screen, 10, 100, 630, 140);
   NewPrintf(screen, 15, 110, "Name:");
   GetString(1, 40, 88, 110, level->name);
}


/*
   Gets the filename.
*/
void GetFilename(char *prompt, char *fname)
{
   Draw3DBox(screen, 10, 100, 630, 140);
   NewPrintf(screen, 15, 110, prompt);
   GetString(1, 30, 20 + NewStrlen(prompt), 110, fname);
}

/*
   Puts a power pill at a certain location.
*/
void PlacePowerPill(LEVEL_DATA *data, int x, int y)
{
   int
      i,
      done = 0;

   for(i = 0; !done && i < 6; i++)
   {
      if(data->power_pills_pos[i].x == -1)
      {
         done = 1;
         data->power_pills_pos[i].x = x;
         data->power_pills_pos[i].y = y;
      }
   }
}

/*
   Picks the tile to place.

   BLOCK1 is first tile by number value.
   BLOCK9 is last tile by number value.
*/
int PickTile(int cur_tile)
{
   int
      cur = cur_tile,
      abort_tile = cur_tile,
      done = 0,
      start = BLOCK1,
      end = BLOCK9;
   BITMAP
      *bmp;

   Draw3DBox(screen, 297, 217, 343, 263);

   while(!done)
   {
      clear_keybuf();

      bmp = main_data[cur].dat;
      blit(bmp, screen, 0, 0, 304, 224, 32, 32);

      if(key[KEY_ESC])  /* abort selection */
      {
         done = 1;
         cur = abort_tile;
      } else if(key[KEY_ENTER])  /* accept tile */
      {
         done = 1;
      } else if(key[KEY_LEFT] || key[KEY_DOWN])  /* previous tile */
      {
         if(cur > start)
         {
            cur--;
         }
      } else if(key[KEY_RIGHT] || key[KEY_UP]) /* next tile */
      {
         if(cur < end)
         {
            cur++;
         }
      }
   }
   clear_keybuf();
   return cur;
}
