#include <stdio.h>
#include <string.h>
#include <conio.h>
#include <stdlib.h>
#include <time.h>
#include <allegro.h>

#include "defs.h"
#include "globals.h"
#include "util.h"
#include "text.h"
#include "edit.h"
#include "pacsnd.h"

#define DEFAULT_LIVES     3
#define MAX_LEVEL        10

/* The states that pacman can be in: */
#define STATE_NORMAL      1    /* moving around */
#define STATE_EATGHOST    2    /* pacman can eat ghosts */
#define STATE_DEAD        3    /* pacman is dead */
#define STATE_DONE        4    /* The level has been completed */
#define STATE_START       5    /* Start the next level. */
#define STATE_COMPLETE    6    /* Completed the game. */
#define STATE_QUIT_DOS    7    /* User is quiting */
#define STATE_QUIT_MENU   8
#define STATE_MENU        9    /* At the main menu */

/* Directions that can be travelled */
#define NOWHERE      0
#define NORTH        1
#define EAST         2
#define SOUTH        3
#define WEST         4

#define GHOST_ALIVE    -1


DATAFILE
   *main_data,
   *text1_data,
   *sound_data;
BITMAP
   *screenBuffer;
int
   edit_mode,
   no_sound,
   use_sound,    /* Is edit mode on? */
   god_mode;
volatile int
   credit_cntr,
   action_cntr,
   total_level_time;
UPDATE_RECT
   *head_rect,
   *last_rect;  /* Just points at last item in list. */

void Initialize(void);
int LoadGameData(void);
void UnloadData(void);
void ShutDown(void);
void MainLoop(void);
void AddUpdateRect(int type, BITMAP *sprite, int x1, int y1, int x2, int y2);
void DeleteUpdateList(void);
int MainMenu(void);
void Hiscores(void);
void VictoryScreen(void);


int main(int argc, char *argv[])
{
   printf(VERSION_INFO);
   fflush(stdout);

   DoCommandArgs(argc, argv);

   Initialize();

   if(edit_mode)
   {
      EditLevel();
   } else
   {
      MainLoop();
   }

   return 0;
}

void CreditTimer(void)
{
   credit_cntr++;
}
END_OF_FUNCTION(CreditTimer);

void ActionTimer(void)
{
   action_cntr++;
   total_level_time++;
}
END_OF_FUNCTION(ActionTimer);

/*
    Loads the needed information and starts up allegro.
    Also sets up functions to be called when the program ends.
*/
void Initialize(void)
{
   /* Setup functions for exit */
   atexit(ShutDown);

   /* Load game data */
   if(!LoadGameData())
   {
      exit(1);
   }

   /* Startup allegro */
   i_love_bill = TRUE;
   allegro_init();
   install_keyboard();
   install_timer();

   /* Start up sounds. */
   use_sound = 0;
   if(!no_sound)
   {
      printf("Starting up sound card...\n");
      fflush(stdout);
      if(install_sound(DIGI_AUTODETECT, MIDI_AUTODETECT, "") == 0)
      {
         printf("Loading \'PACSND.DAT\'\n");
         fflush(stdout);
         sound_data = load_datafile("PACSND.DAT");
         if(sound_data)
         {
            use_sound = 1;
         }
      } else
      {
         printf("Error starting up sound card.  Sounds will not be used.\n");
         fflush(stdout);
      }
   }

   LOCK_VARIABLE(credit_cntr);
   LOCK_FUNCTION(CreditTimer);
   LOCK_VARIABLE(action_cntr);
   LOCK_FUNCTION(ActionTimer);
   LOCK_VARIABLE(total_level_time);

   set_color_depth(8);

   printf("Creating screen buffer.\n");
   fflush(stdout);
   screenBuffer = create_bitmap(640, 480);
   if(!screenBuffer)
   {
      printf("FATAL ERROR: Cannot create screen buffer!\n");
      fflush(stdout);
      exit(1);
   } else
   {
      clear(screenBuffer);
   }

   printf("Press any key to continue...\n");
   fflush(stdout);
   clear_keybuf();
   while(!keypressed()) { }
   clear_keybuf();

   if(set_gfx_mode(GFX_AUTODETECT, 640, 480, 0, 0) < 0)
   {
      printf("FATAL ERROR: Can't switch to a 640x480 256 color screen!\n");
      allegro_exit();
      exit(1);
   }

   set_palette(main_data[PACPAL1].dat);
}

/*
    Loads the game data.

    Returns 1 if successful, or 0 if not.
*/
int LoadGameData(void)
{
   int
      ret_val = 1;

   printf("Loading \'PACMAN.DAT\'\n");
   fflush(stdout);
   main_data = load_datafile("pacman.dat");
   if(!main_data)
   {
      printf("FATAL ERROR:  Can't find \'PACMAN.DAT\'.\n");
      fflush(stdout);
      ret_val = 0;
   }

   printf("Loading \'TEXTDAT1.DAT\'\n");
   fflush(stdout);
   text1_data = load_datafile("textdat1.dat");
   if(!text1_data)
   {
      printf("FATAL ERROR: Can't find \'TEXTDAT1.DAT\'.\n");
      fflush(stdout);
      ret_val = 0;
   }

   return ret_val;
}

/*
   Unloads the game data from memory.
*/
void UnloadData(void)
{
   if(main_data)
   {
      unload_datafile(main_data);
   }
   if(text1_data)
   {
      unload_datafile(text1_data);
   }
   if(sound_data)
   {
      unload_datafile(sound_data);
   }
}

/*
    Frees memory and shuts down program.
*/
void ShutDown(void)
{
   UnloadData();

   if(screenBuffer)
   {
      destroy_bitmap(screenBuffer);
   }
}


/*
   The main loop of the game.
*/
void MainLoop(void)
{
   int
      cur_level = 0,
      lives = DEFAULT_LIVES,
      score,
      old_score,
      x1,
      x2,
      y1,
      y2,
      len,
      tmp_int,
      frame_dir = 0,
      num_food = 0,
      num_power_pills = 0,
      timer_at,
      cnt,
      ghost_eat_timer,
      l_minutes,
      l_seconds,
      score_update;
   LEVEL_DATA
      lev_data;
   UPDATE_RECT
      *cur,
      *tmp;
   PLAYER
      mr_pacman,
      ghosts[4];
   int
      ghosts_deadcnt[4];
   char
      lev_fname[81],
      tmp_string[81];
   BITMAP
      *bmp;

   srandom((int)time(0));

   mr_pacman.state = STATE_START;

   /* Menu loop with game loop inside. */
   while(mr_pacman.state != STATE_QUIT_DOS)
   {
      if(use_sound)
      {
         if(midi_pos != -1)  /* music playing */
         {
            stop_midi();
         }
      }

      mr_pacman.state = MainMenu();

      score = 0;
      old_score = -1;
      lives = DEFAULT_LIVES;
      cur_level = 0;
      ghost_eat_timer = 0;
      score_update = 0;
      clear(screenBuffer);

      // install_int(ActionTimer, 33);  /* 30 frames a sec */
      install_int(ActionTimer, 50);  /* 20 frames a sec */
      clear_keybuf();
      action_cntr = 0;

      /* Main game loop */
      while(mr_pacman.state != STATE_QUIT_DOS &&
            mr_pacman.state != STATE_QUIT_MENU &&
            mr_pacman.state != STATE_COMPLETE &&
            mr_pacman.state != STATE_DEAD)
      {

         if(mr_pacman.state == STATE_START)
         {
            if(use_sound)
            {
               if(midi_pos != -1)  /* music playing */
               {
                  stop_midi();
               }
            }

            cur_level++;
            sprintf(lev_fname, "level%d.lev", cur_level);
            LoadLevel(lev_fname, &lev_data);

            num_food = CountFood(&lev_data);
            num_power_pills = CountPowerPills(&lev_data);

            score_update = 1;
            frame_dir = 1;
            ghost_eat_timer = 0;

            /* Start up pacman data */
            mr_pacman.x = lev_data.pacman_start_pos.x;
            mr_pacman.y = lev_data.pacman_start_pos.y;
            mr_pacman.x_offset = 0;
            mr_pacman.y_offset = 0;
            mr_pacman.direction = EAST;
            mr_pacman.next_dir = NOWHERE;
            mr_pacman.cur_frame = 0;
            mr_pacman.state = STATE_NORMAL;

            /* Start up ghost postions */
            for(cnt = 0; cnt < 4; cnt++)
            {
               ghosts[cnt].x = lev_data.ghost_start_pos.x;
               ghosts[cnt].y = lev_data.ghost_start_pos.y;
               ghosts[cnt].x_offset = 0;
               ghosts[cnt].y_offset = 0;
               ghosts[cnt].direction = (random() & 3) + 1;
               ghosts[cnt].next_dir = NOWHERE;
               ghosts[cnt].cur_frame = 0;
               ghosts[cnt].state = STATE_NORMAL;
               ghosts_deadcnt[cnt] = GHOST_ALIVE;  /* Ghost is alive */
            }

            /* Perform other level startup routines */
            RenderTiles(&lev_data, 0, 0, 20, 14);
            RenderFood(&lev_data, 0, 0, 20, 14);
            RenderPowerPills(&lev_data, 0, 0, 20, 14);

            Draw3DBox(screenBuffer, 10, 185, 630, 295);
            sprintf(tmp_string, "Level #%d", cur_level);
            len = NewStrlen(tmp_string);
            NewPrintf(screenBuffer, (320 - (len >> 1)), 195, tmp_string);
            NewPrintf(screenBuffer, (320-(NewStrlen(lev_data.name)>>1)),
                      225, lev_data.name);
            NewPrintf(screenBuffer,320-(NewStrlen("Press any key...")>>1),
                      255, "Press any key...");
            UpdateScreen();

            clear(screenBuffer);

            clear_keybuf();
            readkey();
            clear_keybuf();

            /* Erase the 3D Box */
            RenderTiles(&lev_data, 0, 0, 20, 14);
            RenderFood(&lev_data, 0, 0, 20, 14);
            RenderPowerPills(&lev_data, 0, 0, 20, 14);

            AddUpdateRect(UPDATE_BACK, NULL, 0, 0, 20, 14);

            old_score = -1;

            if(use_sound)
            {
               play_midi(sound_data[MIDI_LEVEL01 +(cur_level - 1)].dat,
                         TRUE);
            }

            total_level_time = 0;
         }

         /* Draw ghosts */
         cnt = 0;
         do
         {
            if(ghosts_deadcnt[cnt] == GHOST_ALIVE)  /* if ghost is alive */
            {
               x1 = ghosts[cnt].x - 1;
               x2 = ghosts[cnt].x + 2;
               y1 = ghosts[cnt].y - 1;
               y2 = ghosts[cnt].y + 2;

               AddUpdateRect(UPDATE_BACK, NULL, x1, y1, x2, y2);
               bmp = main_data[GHOST1 + cnt].dat;
               AddUpdateRect(UPDATE_SPRITE, bmp,
                         ((ghosts[cnt].x << 5) + ghosts[cnt].x_offset),
                         ((ghosts[cnt].y << 5) + ghosts[cnt].y_offset), 0, 0);
            }
            cnt++;
         } while(cnt < 4);

         /* Draw pacman. */
         x1 = mr_pacman.x - 1;
         x2 = mr_pacman.x + 2;
         y1 = mr_pacman.y - 1;
         y2 = mr_pacman.y + 2;

         AddUpdateRect(UPDATE_BACK, NULL, x1, y1, x2, y2);
         if(mr_pacman.direction == NORTH)
         {
            tmp_int = (ghost_eat_timer) ? PACP4_1 : PAC4_1;
            bmp = main_data[tmp_int + mr_pacman.cur_frame].dat;
         } else if(mr_pacman.direction == EAST)
         {
            tmp_int = (ghost_eat_timer) ? PACP1_1 : PAC1_1;
            bmp = main_data[tmp_int + mr_pacman.cur_frame].dat;
         } else if(mr_pacman.direction == SOUTH)
         {
            tmp_int = (ghost_eat_timer) ? PACP2_1 : PAC2_1;
            bmp = main_data[tmp_int + mr_pacman.cur_frame].dat;
         } else if(mr_pacman.direction == WEST)
         {
            tmp_int = (ghost_eat_timer) ? PACP3_1 : PAC3_1;
            bmp = main_data[tmp_int + mr_pacman.cur_frame].dat;
         } else
         {
            tmp_int = (ghost_eat_timer) ? PACP1_1 : PAC1_1;
            bmp = main_data[tmp_int].dat;
         }

         AddUpdateRect(UPDATE_SPRITE, bmp,
                      ((mr_pacman.x << 5) + mr_pacman.x_offset),
                      ((mr_pacman.y << 5) + mr_pacman.y_offset), 0, 0);


         /* ############# Update Display ################ */
         if(head_rect)   /* any part of the screen to update? */
         {
            /* UPDATE SCREENBUFFER */
            /* update background on screen buffer. */
            cur = head_rect;
            while(cur)
            {
               tmp = cur->next;

               if(cur->type == UPDATE_BACK)
               {
                  x1 = cur->x_top;
                  x2 = cur->x_bottom;
                  y1 = cur->y_top;
                  y2 = cur->y_bottom;
                  RenderTiles(&lev_data, x1, y1, x2, y2);
                  RenderFood(&lev_data, x1, y1, x2, y2);
                  RenderPowerPills(&lev_data, x1, y1, x2, y2);
               }

               cur = tmp;
            }
            /* update sprites on screen buffer. */
            cur = head_rect;
            while(cur)
            {
               tmp = cur->next;

               if(cur->type == UPDATE_SPRITE)
               {
                  x1 = cur->x_top;
                  y1 = cur->y_top;
                  draw_sprite(screenBuffer, cur->sprite, x1, y1);
               }

               cur = tmp;
            }

            /* UPDATE ACTUAL SCREEN USING SCREENBUFFER */
            /* Update background screen */
            cur = head_rect;
            while(cur)
            {
               tmp = cur->next;

               if(cur->type == UPDATE_BACK)
               {
                  x1 = cur->x_top << 5;
                  y1 = cur->y_top << 5;
                  x2 = cur->x_bottom << 5;
                  y2 = cur->y_bottom << 5;
                  blit(screenBuffer, screen, x1, y1, x1, y1,
                       (x2 - x1), (y2 - y1));
               }

               cur = tmp;
            }

            /* Update sprites */
            cur = head_rect;
            while(cur)
            {
               tmp = cur->next;

               if(cur->type == UPDATE_SPRITE)
               {
                  x1 = cur->x_top;
                  y1 = cur->y_top;
                  x2 = cur->x_top + cur->sprite->w;
                  y2 = cur->y_top + cur->sprite->h;
                  blit(screenBuffer, screen, x1, y1, x1, y1,
                       (x2 - x1), (y2 - y1));
               }

               cur = tmp;
            }

            /* Delete list now that it is done. */
            DeleteUpdateList();
         }

         /* ############### Check Inputs ############### */
         if(action_cntr >= 1) /* check keys once every 1/30 of a sec */
         {
            timer_at = action_cntr;

            /* Respond to input */
            if(key[KEY_ESC])  /* Quit */
            {
               if(YesNo("You really want to quit? [Y/N]"))
               {
                  if(use_sound)
                  {
                     play_sample(sound_data[SAMP_DIE].dat, 255,
                                 128, 1000, FALSE);
                  }
                  mr_pacman.state = STATE_QUIT_MENU;
               } else
               {
                  AddUpdateRect(UPDATE_BACK, NULL, 0, 6, 20, 9);
               }
            } else if(key[KEY_UP])
            {
               mr_pacman.next_dir = NORTH;
            } else if(key[KEY_DOWN])
            {
               mr_pacman.next_dir = SOUTH;
            } else if(key[KEY_LEFT])
            {
               mr_pacman.next_dir = WEST;
            } else if(key[KEY_RIGHT])
            {
               mr_pacman.next_dir = EAST;
            }
   
            clear_keybuf();


            /* ########## Perform data updates ############ */

            if(num_food == 0)
            {
               mr_pacman.state = STATE_START;
               tmp_int = total_level_time;
               if(use_sound)
               {
                  play_sample(sound_data[SAMP_DONELEVEL].dat, 255,
                              128, 1000, FALSE);
               }
               /* Display some stats */
               Draw3DBox(screenBuffer, 160, 120, 480, 360);
               sprintf(tmp_string, "Level %d Complete!", cur_level);
               NewPrintf(screenBuffer, 320 - (NewStrlen(tmp_string) >> 1),
                         140, "%s", tmp_string);

               l_minutes = (tmp_int / 20) / 60;
               l_seconds = (tmp_int - (l_minutes * 1200)) / 20;
               NewPrintf(screenBuffer, 180, 170, "Total Time:  %d:%02d",
                         l_minutes, l_seconds);

               NewPrintf(screenBuffer, 180, 200, "Lives remaining: %d",lives);
               UpdateScreen();

               clear_keybuf();
               readkey();
               clear_keybuf();

               if(cur_level == MAX_LEVEL)
               {
                  mr_pacman.state = STATE_COMPLETE;  /* completed the game */
               }
            }

            if(ghost_eat_timer)
            {
               ghost_eat_timer--;
            }

            /* Check to see if pacman and ghost collide. */
            cnt = 0;
            do
            {
               if(ghosts[cnt].x == mr_pacman.x &&
                  ghosts[cnt].y == mr_pacman.y &&
                  ghosts_deadcnt[cnt] == GHOST_ALIVE)  /* they collide */
               {
                  if(ghost_eat_timer) /* Pacman can eat ghosts */
                  {
                     ghosts_deadcnt[cnt] = 200;  /* 10 secs. to respawn */
                     score += 500;
                     if(use_sound)
                     {
                        play_sample(sound_data[SAMP_EATGHOST].dat, 255,
                                 128, 1000, FALSE);
                     }
                  } else   /* ghost kill pacman. */
                  {
                     if(!god_mode)
                     {
                        frame_dir = 1;
                        ghost_eat_timer = 0;
                        mr_pacman.x = lev_data.pacman_start_pos.x;
                        mr_pacman.y = lev_data.pacman_start_pos.y;
                        mr_pacman.x_offset = 0;
                        mr_pacman.y_offset = 0;
                        mr_pacman.direction = EAST;
                        mr_pacman.next_dir = NOWHERE;
                        mr_pacman.cur_frame = 0;
                        score_update = 1;
                        lives--;
                        if(lives < 0)
                        {
                           mr_pacman.state = STATE_DEAD;
                           if(use_sound)
                           {
                              play_sample(sound_data[SAMP_DIE].dat, 255,
                                       128, 1000, FALSE);
                           }
                        } else
                        {
                           mr_pacman.state = STATE_NORMAL;
                           if(use_sound)
                           {
                              play_sample(sound_data[SAMP_GHOSTEAT].dat, 255,
                                       128, 1000, FALSE);
                           }
                        }
                     }
                  }
               }
               cnt++;
            } while(cnt < 4);

            /* Respawn dead ghosts */
            cnt = 0;
            do
            {
               if(ghosts_deadcnt[cnt] != GHOST_ALIVE)  /* ghost is dead */
               {
                  ghosts_deadcnt[cnt]--;
                  if(ghosts_deadcnt[cnt] == GHOST_ALIVE)  /* Ghost now alive */
                  {
                     ghosts[cnt].x = lev_data.ghost_start_pos.x;
                     ghosts[cnt].y = lev_data.ghost_start_pos.y;
                     ghosts[cnt].x_offset = 0;
                     ghosts[cnt].y_offset = 0;
                     ghosts[cnt].direction = (random() & 3) + 1;
                     ghosts[cnt].next_dir = NOWHERE;
                     ghosts[cnt].cur_frame = 0;
                     ghosts[cnt].state = STATE_NORMAL;
                  }
               }
               cnt++;
            } while(cnt < 4);

            /* Move ghosts */
            cnt = 0;
            do
            {
               if(ghosts_deadcnt[cnt] == GHOST_ALIVE)  /* ghost is alive */
               {
                  if(ghosts[cnt].direction == NOWHERE) /* pick a direction */
                  {
                     ghosts[cnt].direction = (random() & 3) + 1;
                  } else
                  {
                     if(ghosts[cnt].x_offset == 0 &&
                        ghosts[cnt].y_offset == 0)
                     {
                        /* Test to see if at intersection. */
                        x1 = ghosts[cnt].x;
                        y1 = ghosts[cnt].y;
                        tmp_int = 0;  /* change direction flag */

                        if(ghosts[cnt].direction == NORTH)
                        {
                           if(lev_data.level_data[y1][x1+1] == -1 ||
                              lev_data.level_data[y1][x1-1] == -1)
                           {
                              tmp_int = 1;
                           }
                        } else if(ghosts[cnt].direction == EAST)
                        {
                           if(lev_data.level_data[y1-1][x1] == -1 ||
                              lev_data.level_data[y1+1][x1] == -1)
                           {
                              tmp_int = 1;
                           }
                        } else if(ghosts[cnt].direction == SOUTH)
                        {
                           if(lev_data.level_data[y1][x1+1] == -1 ||
                              lev_data.level_data[y1][x1-1] == -1)
                           {
                              tmp_int = 1;
                           }
                        } else if(ghosts[cnt].direction == WEST)
                        {
                           if(lev_data.level_data[y1-1][x1] == -1 ||
                              lev_data.level_data[y1+1][x1] == -1)
                           {
                              tmp_int = 1;
                           }
                        }

                        if(tmp_int == 1) /* change direction */
                        {
                           ghosts[cnt].direction = (random() & 3) + 1;
                        }
                     }

                     /* Continue moving on current grid location */
                     tmp_int = ghosts[cnt].direction;
                     x1 = ghosts[cnt].x;
                     y1 = ghosts[cnt].y;
                     switch(tmp_int)
                     {
                        case NORTH:
                        {
                           if(lev_data.level_data[y1-1][x1] == -1)
                           {
                              ghosts[cnt].y_offset -= 4;
                              if(ghosts[cnt].y_offset < -31)
                              {
                                 ghosts[cnt].y--;
                                 ghosts[cnt].y_offset = 0;
                              }
                           } else /* must change direction */
                           {
                              ghosts[cnt].direction = EAST;
                           }
                           break;
                        }
                        case EAST:
                        {
                           if(lev_data.level_data[y1][x1+1] == -1)
                           {
                              ghosts[cnt].x_offset += 4;
                              if(ghosts[cnt].x_offset > 31)
                              {
                                 ghosts[cnt].x++;
                                 ghosts[cnt].x_offset = 0;
                              }
                           } else /* must change direction */
                           {
                              ghosts[cnt].direction = SOUTH;
                           }
                           break;
                        }
                        case SOUTH:
                        {
                           if(lev_data.level_data[y1+1][x1] == -1)
                           {
                              ghosts[cnt].y_offset += 4;
                              if(ghosts[cnt].y_offset > 31)
                              {
                                 ghosts[cnt].y++;
                                 ghosts[cnt].y_offset = 0;
                              }
                           } else /* must change direction */
                           {
                              ghosts[cnt].direction = WEST;
                           }
                           break;
                        }
                        case WEST:
                        {
                           if(lev_data.level_data[y1][x1-1] == -1)
                           {
                              ghosts[cnt].x_offset -= 4;
                              if(ghosts[cnt].x_offset < -31)
                              {
                                 ghosts[cnt].x--;
                                 ghosts[cnt].x_offset = 0;
                              }
                           } else /* must change direction */
                           {
                              ghosts[cnt].direction = NORTH;
                           }
                           break;
                        }
                     }
                  }
               }
               cnt++;
            } while(cnt < 4);


            /* Handle pacman movements */
            if(mr_pacman.x_offset == 0 && mr_pacman.y_offset == 0)
            {
               /* Check for changed directions */
               if(mr_pacman.next_dir != NOWHERE)
               {
                  mr_pacman.direction = mr_pacman.next_dir;
                  mr_pacman.next_dir = NOWHERE;
               }

               /* Check for food */
               if(lev_data.food_pos[mr_pacman.y][mr_pacman.x])
               {
                  lev_data.food_pos[mr_pacman.y][mr_pacman.x] = 0;
                  num_food--;
                  score += 10;
                  if(use_sound)
                  {
                     play_sample(sound_data[SAMP_EATFOOD].dat, 255,
                                 128, 1000, FALSE);
                  }
               }

               /* Check for power pills */
               if(num_power_pills > 0) /* check for power pills */
               {
                  if(IsPowerPillAt(&lev_data, mr_pacman.x, mr_pacman.y))
                  {
                     /* Got the power pill!!!! */
                     ghost_eat_timer = 200;   /* 10 seconds */
                     score += 50;
                     DeleteObject(&lev_data, mr_pacman.x, mr_pacman.y,
                                  DEL_POWER_PILL);
                     if(use_sound)
                     {
                        play_sample(sound_data[SAMP_EATPOWER1].dat, 255,
                                    128, 1000, FALSE);
                     }
                  }
               }
            }
            if(mr_pacman.direction != NOWHERE)
            {
               mr_pacman.cur_frame += frame_dir;
               if(mr_pacman.cur_frame > 2)
               {
                  frame_dir = -1;
               } else if(mr_pacman.cur_frame < 1)
               {
                  frame_dir = 1;
               }

               tmp_int = mr_pacman.direction;
               x1 = mr_pacman.x;
               y1 = mr_pacman.y;
               switch(tmp_int)
               {
                  case NORTH:
                  {
                     if(lev_data.level_data[y1-1][x1] == -1)
                     {
                        mr_pacman.y_offset -= 4;
                        if(mr_pacman.y_offset < -31)
                        {
                           mr_pacman.y--;
                           mr_pacman.y_offset = 0;
                        }
                     } else /* must change direction */
                     {
                        mr_pacman.direction = EAST;
                     }
                     break;
                  }
                  case EAST:
                  {
                     if(lev_data.level_data[y1][x1+1] == -1)
                     {
                        mr_pacman.x_offset += 4;
                        if(mr_pacman.x_offset > 31)
                        {
                           mr_pacman.x++;
                           mr_pacman.x_offset = 0;
                        }
                     } else /* must change direction */
                     {
                        mr_pacman.direction = SOUTH;
                     }
                     break;
                  }
                  case SOUTH:
                  {
                     if(lev_data.level_data[y1+1][x1] == -1)
                     {
                        mr_pacman.y_offset += 4;
                        if(mr_pacman.y_offset > 31)
                        {
                           mr_pacman.y++;
                           mr_pacman.y_offset = 0;
                        }
                     } else /* must change direction */
                     {
                        mr_pacman.direction = WEST;
                     }
                     break;
                  }
                  case WEST:
                  {
                     if(lev_data.level_data[y1][x1-1] == -1)
                     {
                        mr_pacman.x_offset -= 4;
                        if(mr_pacman.x_offset < -31)
                        {
                           mr_pacman.x--;
                           mr_pacman.x_offset = 0;
                        }
                     } else /* must change direction */
                     {
                        mr_pacman.direction = NORTH;
                     }
                     break;
                  }
               }
            }

            if(score != old_score || score_update)
            {
               score_update = 0;
               if(score >= 10000 && old_score < 10000)
               {
                  lives++;
               } else if(score >= 20000 && old_score < 20000)
               {
                  lives++;
               } else if(score >= 30000 && old_score < 30000)
               {
                  lives++;
               } else if(score >= 40000 && old_score < 40000)
               {
                  lives++;
               } else if(score >= 50000 && old_score < 50000)
               {
                  lives++;
               }
               old_score = score;
               rectfill(screenBuffer, 0, 448, 640, 480, 0);
               NewPrintf(screenBuffer, 10, 460, "Lives: %d", lives);
               NewPrintf(screenBuffer, 150, 460, "Score: %d", score);
               NewPrintf(screenBuffer, 450, 460, "Level: %d", cur_level);
               blit(screenBuffer, screen, 0, 448, 0, 448, 640, 32);
            }

            action_cntr = 0;
         }
      }

      /* Timer not needed now, remove it. */
      remove_int(ActionTimer);

      if(mr_pacman.state == STATE_COMPLETE)  /* completed all the levels */
      {
         if(use_sound)
         {
            stop_midi();

            play_midi(sound_data[MIDI_VICTORY].dat,
                      TRUE);
         }
         VictoryScreen();
         if(use_sound)
         {
            stop_midi();
         }
      }

      /* Add score and name to hiscore list if player made it. */
      if(mr_pacman.state != STATE_QUIT_DOS)
      {
         AddToHiscores(score);
      }
   }
}

/*
   Adds a rectangle describing part of the screen to be updated.
*/
void AddUpdateRect(int type, BITMAP *sprite, int x1, int y1, int x2, int y2)
{
   UPDATE_RECT
      *new_rect;

   new_rect = (UPDATE_RECT *)malloc(sizeof(UPDATE_RECT));

   if(new_rect)
   {
      if(!head_rect)
      {
         head_rect = new_rect;
      } else
      {
         last_rect->next = new_rect;
      }

      last_rect = new_rect;

      new_rect->next = NULL;
      new_rect->type = type;
      new_rect->sprite = sprite;
      new_rect->x_top = x1;
      new_rect->y_top = y1;
      new_rect->x_bottom = x2;
      new_rect->y_bottom = y2;
   }
}


/*
   Deletes the screen update list.
*/
void DeleteUpdateList(void)
{
   UPDATE_RECT
      *cur,
      *tmp;

   cur = head_rect;
   while(cur)
   {
      tmp = cur->next;
      free(cur);
      cur = tmp;
   }

   head_rect = NULL;
   last_rect = NULL;
}


/*
   The main menu that asks the user to run the game,
   view the hiscores or quit.
 */
int MainMenu(void)
{
   int
      ret_state = 0,
      picked_choice = 0,
      pos = 0,             /* the item that is selected */
      update = 1;
   BITMAP
      *bmp;

   while(!picked_choice)
   {
      clear_keybuf();

      /* update screen. */
      if(update)
      {
         update = 0;
         clear(screenBuffer);
      
         bmp = main_data[BIGPAC1].dat;
         blit(bmp, screenBuffer, 0, 0, 220, 0, bmp->w, bmp->h);

         NewPrintf(screenBuffer, 250, 300, "New Game");
         NewPrintf(screenBuffer, 250, 330, "Hiscores");
         NewPrintf(screenBuffer, 250, 360, "Edit Level");
         NewPrintf(screenBuffer, 250, 390, "Credits");
         NewPrintf(screenBuffer, 250, 420, "Quit");

         bmp = main_data[POWER1].dat;
         draw_sprite(screenBuffer, bmp, 215, (300 + (pos * 30)));

         UpdateScreen();
      }

      if(key[KEY_ENTER])
      {
         if(pos == 0)  /* New game */
         {
            ret_state = STATE_START;
            picked_choice = 1;
         } else if(pos == 1)
         {
            update = 1;
            Hiscores();
         } else if(pos == 2)
         {
            update = 1;
            EditLevel();
         } else if(pos == 3)
         {
            update = 1;
            Credits();
         } else if(pos == 4)
         {
            ret_state = STATE_QUIT_DOS;
            picked_choice = 1;
         }
      } else if(key[KEY_UP])
      {
         if(use_sound)
         {
            play_sample(sound_data[SAMP_EATFOOD].dat, 255, 128, 1000, FALSE);
         }
         if(pos == 0)
         {
            pos = 4;
         } else
         {
            pos--;
         }
         update = 1;
      } else if(key[KEY_DOWN])
      {
         if(use_sound)
         {
            play_sample(sound_data[SAMP_EATFOOD].dat, 255, 128, 1000, FALSE);
         }
         if(pos == 4)
         {
            pos = 0;
         } else
         {
            pos++;
         }
         update = 1;
      }
   }

   return ret_state;
}



/*
   Displays the Hiscores
*/
void Hiscores(void)
{
   static char
      *default_score_names[10] = { "Pacman", "Pacman", "Pacman", "Pacman",
                                   "Pacman", "Pacman", "Pacman", "Pacman",
                                   "Pacman", "Pacman"};
   static int
      default_score_numbers[10] = { 5000, 4500, 4000, 3500, 3000,
                                    2500, 2000, 1500, 1000, 500};
   FILE
      *fp;
   int
      x,
      num;
   char
      tmp[81],
      name[41];

   clear(screenBuffer);

   strcpy(tmp, "*****  HISCORES  *****");
   x = NewStrlen(tmp);
   NewPrintf(screenBuffer, (320 - (x >> 1)), 0, tmp);

   fp = fopen("hiscore.dat", "r");
   if(fp)
   {
      for(x = 0; x < 10 && !feof(fp); x++)
      {
         fscanf(fp, "%40s\t%d\n", name, &num);
         NewPrintf(screenBuffer, 10, (50 + (x * 30)), "%s", name);
         NewPrintf(screenBuffer, 500, (50 + (x * 30)), "%d", num);
      }
      fclose(fp);
   } else /* Can't open file so create it. */
   {
      fp = fopen("hiscore.dat", "w");
      if(fp)
      {
         for(x = 0; x < 10; x++)
         {
            fprintf(fp, "%s\t%d\n", default_score_names[x],
                    default_score_numbers[x]);
            NewPrintf(screenBuffer, 10, (50 + (x * 30)), "%s",
                    default_score_names[x]);
            NewPrintf(screenBuffer, 500, (50 + (x * 30)), "%d",
                    default_score_numbers[x]);
         }
         fclose(fp);
      }
   }

   strcpy(tmp, "Press any key to return...");
   x = NewStrlen(tmp);
   NewPrintf(screenBuffer, 320 - (x >> 1), 450, tmp);

   UpdateScreen();

   clear_keybuf();
   readkey();
   clear_keybuf();
}


/*
   Displays the victory screen.
*/
void VictoryScreen(void)
{
   int
      x,
      frame = 0,
      dir = 1,
      i;
   char
      tmp[128];
   static char
      *victory_txt[] = { "Congradulations!  You just beat",
                        VERSION_INFO,
                        "Now try that again, I just dare ya!",
                        NULL};

   clear(screenBuffer);

   for(i = 0; victory_txt[i]; i++)
   {
      x = NewStrlen(victory_txt[i]);
      NewPrintf(screenBuffer, 320 - (x >> 1), 100 + i * 25, victory_txt[i]);
   }

   strcpy(tmp, "Press any key to return...");
   x = NewStrlen(tmp);
   NewPrintf(screenBuffer, 320 - (x >> 1), 450, tmp);

   UpdateScreen();

   install_int(CreditTimer, 10);
   credit_cntr = 0;

   clear_keybuf();
   while(!keypressed())
   {
      if(credit_cntr > 5)
      {
         credit_cntr = 0;
         for(i = 0; i < 15; i++)
         {
            blit(main_data[PAC1_1+frame].dat, screen, 0, 0, 0, i<<5,32,32);
            blit(main_data[PAC3_1+frame].dat, screen, 0, 0, 608, i<<5,32,32);
         }
         frame += dir;
         if(frame > 3)
         {
            frame = 2;
            dir = -1;
         } else if(frame < 0)
         {
            frame = 0;
            dir = 1;
         }
      }
   }
   remove_int(CreditTimer);
   clear_keybuf();
}

