#include <stdio.h>
#include <string.h>

#include "defs.h"
#include "globals.h"
#include "util.h"
#include "text.h"

/*
   Checks to see if a file exits.
   Returns 1 if it does, or 0 if not.
*/
int FileExist(char *filename)
{
   int
      ret_val = 0;
   FILE
      *fp;

   fp = fopen(filename, "rb");
   if(fp)
   {
      ret_val = 1;
      fclose(fp);
   }

   return ret_val;
}

/*
   Looks for a specific command line argument based on (*look_for).
   If the argument is found, then the array position is returned.
   If the argument is not found, then -1 is returned.
*/
int ParseCommandArgs(int arg_count, char *arguments[], char *look_for)
{
   int
      cnt,
      ret_val = -1,
      done = 0;

   for(cnt = 0; !done && cnt < arg_count; cnt++)
   {
      if(!stricmp(arguments[cnt], look_for))
      {
         done = 1;
         ret_val = cnt;
      }
   }

   return ret_val;
}


/*
   Scrambles a buffer of data.

   Note:  Calling this function on scrambled data will
          unscramble it.

   Returns 0 on success.
*/
int pat_scramble(char *data, unsigned long size)
{
   int
      ret_value = 0;
   unsigned long
      buf_pos = 0L;
   unsigned char
      c;

   while(buf_pos < size)
   {
      /* Swap nibbles */
      c = (data[buf_pos] & 0x0F) << 4;
      c |= (data[buf_pos] & 0xF0) >> 4;
      /* Reverse all bits */
      c ^= 0xFF;
      data[buf_pos++] = c;
   }
   return ret_value;
}

/*
   Checks the command line arguments and takes the
   appropriate action.
*/
void DoCommandArgs(int argc, char *argv[])
{
   int
      index;

   index = ParseCommandArgs(argc, argv, "-edit");   /* Edit mode */
   if(index != -1)
   {
      printf("Edit mode on.\n");
      fflush(stdout);
      edit_mode = 1;
   } else
   {
      edit_mode = 0;
   }

   index = ParseCommandArgs(argc, argv, "-help");
   if(index != -1)
   {
      printf("----------------------------------------------\n");
      printf("|                  HELP                       |\n");
      printf("|---------------------------------------------|\n");
      printf("| -help        = Command line parameter help. |\n");
      printf("| -edit        = Turn edit mode on.           |\n");
      printf("| -nosound     = Turns off all sounds.        |\n");
      printf("-----------------------------------------------\n");
      fflush(stdout);
   }

   index = ParseCommandArgs(argc, argv, "-nosound");
   if(index != -1)
   {
      no_sound = 1;
      printf("Sounds turned off.\n");
      fflush(stdout);
   }

   index = ParseCommandArgs(argc, argv, "-godmode");
   if(index != -1)
   {
      god_mode = 1;
      printf("God mode on.\n");
      fflush(stdout);
   }
}

/*
   Updates the screen with the double buffer.
*/
void UpdateScreen(void)
{
   blit(screenBuffer, screen, 0, 0, 0, 0, 640, 480);
}


void CreditTimer(void);
/*
   Prints the credits on the screen.
*/
void Credits(void)
{
   int
      x,
      frame = 0,
      dir = 1,
      i;
   char
      tmp[128];
   static char
      *credit_txt[] = { VERSION_INFO,
                        "Programming:  Patrick Barry",
                        NULL};

   clear(screenBuffer);

   for(i = 0; credit_txt[i]; i++)
   {
      x = NewStrlen(credit_txt[i]);
      NewPrintf(screenBuffer, 320 - (x >> 1), i * 25, credit_txt[i]);
   }

   strcpy(tmp, "Press any key to return...");
   x = NewStrlen(tmp);
   NewPrintf(screenBuffer, 320 - (x >> 1), 450, tmp);

   UpdateScreen();

   install_int(CreditTimer, 10);
   credit_cntr = 0;

   clear_keybuf();
   while(!keypressed())
   {
      if(credit_cntr > 5)
      {
         credit_cntr = 0;
         blit(main_data[PAC1_1+frame].dat, screen, 0, 0, 0, 0, 32, 32);
         blit(main_data[PAC3_1+frame].dat, screen, 0, 0, 608, 0 , 32, 32);
         blit(main_data[PAC1_1+frame].dat, screen, 0, 0, 0, 448, 32, 32);
         blit(main_data[PAC3_1+frame].dat, screen, 0, 0, 608, 448, 32, 32);
         frame += dir;
         if(frame > 3)
         {
            frame = 2;
            dir = -1;
         } else if(frame < 0)
         {
            frame = 0;
            dir = 1;
         }
      }
   }
   remove_int(CreditTimer);
   clear_keybuf();
}


/*
   Draws the wall tiles.
*/
void RenderTiles(LEVEL_DATA *data, int xs, int ys, int xe, int ye)
{
   int
      x,
      y,
      tile;

   y = ys;
   do
   {
      x = xs;
      do
      {
         tile = data->level_data[y][x];
         if(tile >= 0)
         {
            blit(main_data[tile].dat, screenBuffer, 0, 0,
                 (x << 5), (y << 5), 32, 32);
         } else  /* Black background */
         {
            rectfill(screenBuffer, (x << 5), (y << 5),
                     ((x << 5) + 32), ((y << 5) + 32), 0);
         }
         x++;
      } while(x < xe);
      y++;
   } while(y < ye);
}

/*
   Draws a 3d Box.
*/
void Draw3DBox(BITMAP *bmp, int x1, int y1, int x2, int y2)
{
   rectfill(bmp, x1, y1, x2, y2, 0);

   hline(bmp, x1, y1, x2, 80);
   hline(bmp, x1, y1 + 1, x2 - 1, 80);
   vline(bmp, x1, y1, y2, 80);
   vline(bmp, x1 + 1, y1, y2 - 1, 80);

   hline(bmp, x1 + 1, y2 - 1, x2, 100);
   hline(bmp, x1, y2, x2, 100);
   vline(bmp, x2, y1 + 1, y2, 100);
   vline(bmp, x2 - 1, y1 + 2, y2, 100);
}


void RenderFood(LEVEL_DATA *data, int xs, int ys, int xe, int ye)
{
   int
      x,
      y,
      x_off,
      y_off;
   BITMAP
      *bmp;

   bmp = main_data[FOOD1].dat;
   x_off = 16 - (bmp->w >> 1);
   y_off = 16 - (bmp->h >> 1);

   y = ys;
   do
   {
      x = xs;
      do
      {
         if(data->food_pos[y][x])
         {
            blit(bmp, screenBuffer, 0, 0,
                 ((x << 5) + x_off), ((y << 5) + y_off), 32, 32);
         }
         x++;
      } while(x < xe);
      y++;
   } while(y < ye);
}

/*
   Removes all objects located at (x, y).
*/
void DeleteObject(LEVEL_DATA *data, int x, int y, int type)
{
   int
      i;

   if(type == DEL_FOOD1 || type == DEL_ALL)
   {
      data->food_pos[y][x] = 0;
   }
   if(type == DEL_GHOST_START || type == DEL_ALL)
   {
      if(data->ghost_start_pos.x == x && data->ghost_start_pos.y == y)
      {
         data->ghost_start_pos.x = -1;
         data->ghost_start_pos.y = -1;
      }
   }
   if(type == DEL_PACMAN_START || type == DEL_ALL)
   {
      if(data->pacman_start_pos.x == x && data->pacman_start_pos.y == y)
      {
         data->pacman_start_pos.x = -1;
         data->pacman_start_pos.y = -1;
      }
   }
   if(type == DEL_POWER_PILL || type == DEL_ALL)
   {
      for(i = 0; i < 6; i++)
      {
         if(data->power_pills_pos[i].x == x &&
            data->power_pills_pos[i].y == y)
         {
            data->power_pills_pos[i].x = -1;
            data->power_pills_pos[i].y = -1;
         }
      }
   }
   if(type == DEL_TILES || type == DEL_ALL)
   {
      data->level_data[y][x] = -1;
   }
}

/*
   Draws the power pills.
*/
void RenderPowerPills(LEVEL_DATA *data, int xs, int ys, int xe, int ye)
{
   int
      x,
      y,
      x_off,
      y_off,
      i;
   BITMAP
      *bmp;

   bmp = main_data[POWER1].dat;
   x_off = 16 - (bmp->w >> 1);
   y_off = 16 - (bmp->h >> 1);

   for(i = 0; i < 6; i++)
   {
      x = data->power_pills_pos[i].x;
      y = data->power_pills_pos[i].y;
      if((x >= xs && x <= xe) && (y >= ys && y <= ye))
      {
         blit(main_data[POWER1].dat, screenBuffer, 0, 0,
             ((x << 5) + x_off), ((y << 5) + y_off), 32, 32);
      }
   }
}

/*
   Counts the number of food pelletes in a level.
*/
int CountFood(LEVEL_DATA *data)
{
   int
      x,
      y,
      cnt = 0;

   for(y = 0; y < 14; y++)
   {
      for(x = 0; x < 20; x++)
      {
         if(data->food_pos[y][x])
         {
            cnt++;
         }
      }
   }

   return cnt;
}

/*
   Counts the number of power pills in a level.
*/
int CountPowerPills(LEVEL_DATA *data)
{
   int
      cnt = 0,
      i;

   for(i = 0; i < 6; i++)
   {
      if(data->power_pills_pos[i].x > -1 &&
         data->power_pills_pos[i].y > -1)
      {
         cnt++;
      }
   }

   return cnt;
}

/*
   Swaps the contents of two integers.
*/
void SwapInt(int *a, int *b)
{
   int
      c;

   c = *a;
   *a = *b;
   *a = c;
}


/*
    Presents the user with a prompt that asks a yes/no question.
*/
int YesNo(char *prompt)
{
   int
      ret_val = 0,
      done = 0,
      k,
      len,
      x1,
      x2;

   len = NewStrlen(prompt);
   x1 = 320 - (len >> 1);
   x2 = 330 + (len >> 1);

   Draw3DBox(screenBuffer, (x1 - 10), 220, x2, 260);
   NewPrintf(screenBuffer, x1, 230, "%s", prompt);
   UpdateScreen();

   while(!done)
   {
      clear_keybuf();
      k = readkey() & 0x00FF;

      if(k == 'n' || k == 'N')
      {
         done = 1;
         ret_val = 0;
      } else if(k == 'y' || k == 'K')
      {
         done = 1;
         ret_val = 1;
      }
   }

   return ret_val;
}

/*
   Returns 1 if a power pill is at the specified location, or 0 if not.
*/
int IsPowerPillAt(LEVEL_DATA *data, int x, int y)
{
   int
      cnt,
      ret_val = 0;

   cnt = 0;
   do
   {
      if(data->power_pills_pos[cnt].x == x &&
         data->power_pills_pos[cnt].y == y)
      {
         ret_val = 1;
      }
      cnt++;
   } while(cnt < 6 && !ret_val);

   return ret_val;
}

/*
    If the given score can be added to the list of hiscores then the
    name of the player is asked for.  Otherwise the list remains unchaged.
*/
extern void Hiscores(void);  /* located in PACMAN.C */
void AddToHiscores(int sc)
{
   int
      scores[10],
      cnt,
      move_from = 0,
      changed = 0;
   char
      names[10][41],
      player_name[41];
   FILE
      *fp;

   fp = fopen("hiscore.dat", "r");
   if(fp)
   {
      for(cnt = 0; cnt < 10 && !feof(fp); cnt++)
      {
         fscanf(fp, "%40s\t%d\n", names[cnt], &scores[cnt]);
      }
      fclose(fp);

      for(cnt = 0; cnt < 10 && !changed; cnt++)
      {
         if(sc > scores[cnt])
         {
            move_from = cnt;
            changed = 1;
         }
      }

      if(changed)  /* made it into score list */
      {
         /* Get players name */
         player_name[0] = 0;
         Draw3DBox(screen, 10, 100, 630, 140);
         NewPrintf(screen, 15, 110, "Name:");
         GetString(1, 40, 88, 110, player_name);

         /* Insert score and name into list. */
         for(cnt = 9; cnt > (move_from - 1); cnt--)
         {
            scores[cnt] = scores[cnt - 1];
            strcpy(names[cnt], names[cnt - 1]);
         }
         scores[move_from] = sc;
         strcpy(names[move_from], player_name);

         /* Update the list */
         fp = fopen("hiscore.dat", "w");
         if(fp)
         {
            for(cnt = 0; cnt < 10; cnt++)
            {
               fprintf(fp, "%s\t%d\n", names[cnt], scores[cnt]);
            }
            fclose(fp);
         }
      }
   }

   Hiscores();
}


